from pydantic import BaseModel


class StatusSummary(BaseModel):
    count: int = 0
    case_ids: list[str] = []


class AnalysisSummary(BaseModel):
    order_id: int
    cancelled: StatusSummary
    completed: StatusSummary
    delivered: StatusSummary
    failed: StatusSummary
    running: StatusSummary

    @property
    def case_ids(self) -> list[str]:
        cancelled_case_ids: list[str] = self.cancelled.case_ids
        completed_case_ids: list[str] = self.completed.case_ids
        delivered_case_ids: list[str] = self.delivered.case_ids
        failed_case_ids: list[str] = self.failed.case_ids
        running_case_ids: list[str] = self.running.case_ids
        return (
            cancelled_case_ids
            + completed_case_ids
            + delivered_case_ids
            + failed_case_ids
            + running_case_ids
        )


class SummariesResponse(BaseModel):
    summaries: list[AnalysisSummary]



from enum import StrEnum
from pydantic import BaseModel, Field


class OrderSortField(StrEnum):
    ORDER_DATE: str = "order_date"
    ID: str = "id"
    CUSTOMER_ID: str = "customer_id"


class SortOrder(StrEnum):
    ASC: str = "asc"
    DESC: str = "desc"


class OrdersRequest(BaseModel):
    page_size: int | None = Field(alias="pageSize", default=50)
    page: int | None = 1
    sort_field: OrderSortField | None = Field(alias="sortField", default=OrderSortField.ORDER_DATE)
    sort_order: SortOrder | None = Field(alias="sortOrder", default=SortOrder.DESC)
    search: str | None = None
    workflow: str | None = None
    is_open: bool | None = None



from pathlib import Path
from pydantic import BaseModel

from cg.constants.constants import ReadDirection


class FastqFile(BaseModel):
    """A fastq file with a sample name, file path and read direction."""

    sample_name: str
    fastq_file_path: Path
    read_direction: ReadDirection



"""Util functions that provide support in mapping objects."""

import re

from cg.exc import CgError


def get_item_by_pattern_in_source(source: str, pattern_map: dict[str, any]) -> any:
    """
    Check if the pattern is in the keys of the pattern map.
    Raises an CgError.
    """
    for map_key in pattern_map.keys():
        if re.search(map_key, source):
            return pattern_map.get(map_key)
    raise CgError(f"Could not find pattern for {source} in {pattern_map.keys()}.")



from cg.models.orders.sample_base import StatusEnum
from cg.services.orders.validation.errors.case_errors import (
    CaseDoesNotExistError,
    CaseNameNotAvailableError,
    CaseOutsideOfCollaborationError,
    DoubleNormalError,
    DoubleTumourError,
    ExistingCaseWithoutAffectedSampleError,
    MoreThanTwoSamplesInCaseError,
    MultiplePrepCategoriesError,
    MultipleSamplesInCaseError,
    NewCaseWithoutAffectedSampleError,
    NormalOnlyWGSError,
    NumberOfNormalSamplesError,
    RepeatedCaseNameError,
    RepeatedGenePanelsError,
)
from cg.services.orders.validation.models.case import Case
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.balsamic.models.order import BalsamicOrder
from cg.services.orders.validation.order_types.balsamic_umi.models.order import BalsamicUmiOrder
from cg.services.orders.validation.order_types.mip_dna.models.order import MIPDNAOrder
from cg.services.orders.validation.order_types.nallo.models.order import NalloOrder
from cg.services.orders.validation.rules.case.utils import (
    contains_duplicates,
    get_case_prep_categories,
    is_case_not_from_collaboration,
    is_double_normal,
    is_double_tumour,
    is_normal_only_wgs,
)
from cg.services.orders.validation.rules.case_sample.utils import get_repeated_case_name_errors
from cg.store.models import Case as DbCase
from cg.store.store import Store


def validate_gene_panels_unique(order: OrderWithCases, **kwargs) -> list[RepeatedGenePanelsError]:
    errors: list[RepeatedGenePanelsError] = []
    for case_index, case in order.enumerated_new_cases:
        if contains_duplicates(case.panels):
            error = RepeatedGenePanelsError(case_index=case_index)
            errors.append(error)
    return errors


def validate_case_names_available(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[CaseNameNotAvailableError]:
    errors: list[CaseNameNotAvailableError] = []
    customer = store.get_customer_by_internal_id(order.customer)
    for case_index, case in order.enumerated_new_cases:
        if store.get_case_by_name_and_customer(case_name=case.name, customer=customer):
            error = CaseNameNotAvailableError(case_index=case_index)
            errors.append(error)
    return errors


def validate_case_internal_ids_exist(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[CaseDoesNotExistError]:
    errors: list[CaseDoesNotExistError] = []
    for case_index, case in order.enumerated_existing_cases:
        case: Case | None = store.get_case_by_internal_id(case.internal_id)
        if not case:
            error = CaseDoesNotExistError(case_index=case_index)
            errors.append(error)
    return errors


def validate_existing_cases_belong_to_collaboration(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[CaseOutsideOfCollaborationError]:
    """Validates that all existing cases within the order belong to a customer
    within the order's customer's collaboration."""
    errors: list[CaseOutsideOfCollaborationError] = []
    for case_index, case in order.enumerated_existing_cases:
        if is_case_not_from_collaboration(case=case, customer_id=order.customer, store=store):
            error = CaseOutsideOfCollaborationError(case_index=case_index)
            errors.append(error)
    return errors


def validate_case_names_not_repeated(
    order: OrderWithCases,
    **kwargs,
) -> list[RepeatedCaseNameError]:
    return get_repeated_case_name_errors(order)


def validate_one_sample_per_case(
    order: OrderWithCases, **kwargs
) -> list[MultipleSamplesInCaseError]:
    """Validates that there is only one sample in each case.
    Only applicable to RNAFusion."""
    errors: list[MultipleSamplesInCaseError] = []
    for case_index, case in order.enumerated_new_cases:
        if len(case.samples) > 1:
            error = MultipleSamplesInCaseError(case_index=case_index)
            errors.append(error)
    return errors


def validate_at_most_two_samples_per_case(
    order: OrderWithCases, **kwargs
) -> list[MoreThanTwoSamplesInCaseError]:
    """Validates that there is at most two samples in each case.
    Only applicable to Balsamic and Balsamic-UMI."""
    errors: list[MoreThanTwoSamplesInCaseError] = []
    for case_index, case in order.enumerated_new_cases:
        if len(case.samples) > 2:
            error = MoreThanTwoSamplesInCaseError(case_index=case_index)
            errors.append(error)
    return errors


def validate_number_of_normal_samples(
    order: BalsamicOrder | BalsamicUmiOrder, store: Store, **kwargs
) -> list[NumberOfNormalSamplesError]:
    """
    Validates that Balsamic cases with pairs of samples contain one tumour and one normal sample
    and that cases with one WGS sample only contain a tumour sample.
    Only applicable to Balsamic and Balsamic-UMI.
    """
    errors: list[NumberOfNormalSamplesError] = []
    for case_index, case in order.enumerated_new_cases:
        if is_double_normal(case=case, store=store):
            error = DoubleNormalError(case_index=case_index)
            errors.append(error)
        elif is_double_tumour(case=case, store=store):
            error = DoubleTumourError(case_index=case_index)
            errors.append(error)
        if is_normal_only_wgs(case=case, store=store):
            error = NormalOnlyWGSError(case_index=case_index)
            errors.append(error)
    return errors


def validate_each_new_case_has_an_affected_sample(
    order: MIPDNAOrder | NalloOrder, **kwargs
) -> list[NewCaseWithoutAffectedSampleError]:
    """Validates that each case in the order contains at least one sample with affected status."""
    errors: list[NewCaseWithoutAffectedSampleError] = []
    for case_index, case in order.enumerated_new_cases:
        if all(sample.status != StatusEnum.affected for sample in case.samples):
            error = NewCaseWithoutAffectedSampleError(case_index=case_index)
            errors.append(error)
    return errors


def validate_existing_cases_have_an_affected_sample(
    order: MIPDNAOrder | NalloOrder, store: Store, **kwargs
) -> list[ExistingCaseWithoutAffectedSampleError]:
    errors: list[ExistingCaseWithoutAffectedSampleError] = []
    for case_index, case in order.enumerated_existing_cases:
        db_case: DbCase = store.get_case_by_internal_id(case.internal_id)
        if all(link.status != StatusEnum.affected for link in db_case.links):
            error = ExistingCaseWithoutAffectedSampleError(case_index=case_index)
            errors.append(error)
    return errors


def validate_samples_in_case_have_same_prep_category(
    order: OrderWithCases, store: Store, **kwargs
) -> list[MultiplePrepCategoriesError]:
    errors: list[MultiplePrepCategoriesError] = []
    for case_index, case in order.enumerated_new_cases:
        prep_categories: set[str] = get_case_prep_categories(case=case, store=store)
        if len(prep_categories) > 1:
            error = MultiplePrepCategoriesError(case_index=case_index)
            errors.append(error)
    return errors



from collections import Counter

from cg.models.orders.constants import OrderType
from cg.services.orders.validation.constants import ALLOWED_SKIP_RC_BUFFERS
from cg.services.orders.validation.errors.case_errors import InvalidGenePanelsError
from cg.services.orders.validation.errors.case_sample_errors import (
    ApplicationArchivedError,
    ApplicationNotCompatibleError,
    ApplicationNotValidError,
    BufferMissingError,
    CaptureKitMissingError,
    CaptureKitResetError,
    ConcentrationRequiredIfSkipRCError,
    ContainerNameMissingError,
    ContainerNameRepeatedError,
    ExistingSampleWrongTypeError,
    FatherNotInCaseError,
    InvalidBufferError,
    InvalidCaptureKitError,
    InvalidConcentrationIfSkipRCError,
    InvalidFatherSexError,
    InvalidMotherSexError,
    InvalidVolumeError,
    MotherNotInCaseError,
    OccupiedWellError,
    PedigreeError,
    SampleDoesNotExistError,
    SampleNameAlreadyExistsError,
    SampleNameRepeatedError,
    SampleNameSameAsCaseNameError,
    SampleOutsideOfCollaborationError,
    SexSubjectIdError,
    StatusUnknownError,
    SubjectIdSameAsCaseNameError,
    SubjectIdSameAsSampleNameError,
    VolumeRequiredError,
    WellFormatError,
    WellPositionMissingError,
)
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.models.sample_aliases import SampleInCase
from cg.services.orders.validation.order_types.balsamic.models.order import BalsamicOrder
from cg.services.orders.validation.order_types.balsamic_umi.models.order import BalsamicUmiOrder
from cg.services.orders.validation.rules.case_sample.pedigree.validate_pedigree import (
    get_pedigree_errors,
)
from cg.services.orders.validation.rules.case_sample.utils import (
    are_all_samples_unknown,
    get_counter_container_names,
    get_existing_case_names,
    get_existing_sample_names,
    get_father_case_errors,
    get_father_sex_errors,
    get_invalid_panels,
    get_mother_case_errors,
    get_mother_sex_errors,
    get_occupied_well_errors,
    get_well_sample_map,
    has_sex_and_subject,
    is_buffer_missing,
    is_concentration_missing,
    is_container_name_missing,
    is_invalid_plate_well_format,
    is_sample_compatible_with_order_type,
    is_sample_not_from_collaboration,
    is_sample_tube_name_reused,
    is_well_position_missing,
    validate_concentration_in_case,
    validate_subject_ids_in_case,
)
from cg.services.orders.validation.rules.utils import (
    does_sample_need_capture_kit,
    is_application_compatible,
    is_invalid_capture_kit,
    is_sample_missing_capture_kit,
    is_volume_invalid,
    is_volume_missing,
)
from cg.store.models import Sample as DbSample
from cg.store.store import Store


def validate_application_compatibility(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[ApplicationNotCompatibleError]:
    errors: list[ApplicationNotCompatibleError] = []
    order_type: OrderType = order.order_type
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if not is_application_compatible(
                order_type=order_type,
                application_tag=sample.application,
                store=store,
            ):
                error = ApplicationNotCompatibleError(
                    case_index=case_index,
                    sample_index=sample_index,
                )
                errors.append(error)
    return errors


def validate_buffer_skip_rc_condition(order: OrderWithCases, **kwargs) -> list[InvalidBufferError]:
    errors: list[InvalidBufferError] = []
    if order.skip_reception_control:
        errors.extend(validate_buffers_are_allowed(order))
    return errors


def validate_buffers_are_allowed(order: OrderWithCases, **kwargs) -> list[InvalidBufferError]:
    errors: list[InvalidBufferError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if sample.elution_buffer not in ALLOWED_SKIP_RC_BUFFERS:
                error = InvalidBufferError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_concentration_required_if_skip_rc(
    order: OrderWithCases, **kwargs
) -> list[ConcentrationRequiredIfSkipRCError]:
    if not order.skip_reception_control:
        return []
    errors: list[ConcentrationRequiredIfSkipRCError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_concentration_missing(sample):
                error = ConcentrationRequiredIfSkipRCError(
                    case_index=case_index,
                    sample_index=sample_index,
                )
                errors.append(error)
    return errors


def validate_subject_ids_different_from_sample_names(
    order: OrderWithCases, **kwargs
) -> list[SubjectIdSameAsSampleNameError]:
    errors: list[SubjectIdSameAsSampleNameError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if sample.name == sample.subject_id:
                error = SubjectIdSameAsSampleNameError(
                    case_index=case_index,
                    sample_index=sample_index,
                )
                errors.append(error)
    return errors


def validate_well_positions_required(
    order: OrderWithCases, **kwargs
) -> list[WellPositionMissingError]:
    errors: list[WellPositionMissingError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_well_position_missing(sample):
                error = WellPositionMissingError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_container_name_required(
    order: OrderWithCases, **kwargs
) -> list[ContainerNameMissingError]:
    errors: list[ContainerNameMissingError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_container_name_missing(sample):
                error = ContainerNameMissingError(
                    case_index=case_index,
                    sample_index=sample_index,
                )
                errors.append(error)
    return errors


def validate_application_exists(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[ApplicationNotValidError]:
    errors: list[ApplicationNotValidError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if not store.get_application_by_tag(sample.application):
                error = ApplicationNotValidError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_application_not_archived(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[ApplicationArchivedError]:
    errors: list[ApplicationArchivedError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if store.is_application_archived(sample.application):
                error = ApplicationArchivedError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_gene_panels_exist(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[InvalidGenePanelsError]:
    errors: list[InvalidGenePanelsError] = []
    for case_index, case in order.enumerated_new_cases:
        if invalid_panels := get_invalid_panels(panels=case.panels, store=store):
            case_error = InvalidGenePanelsError(case_index=case_index, panels=invalid_panels)
            errors.append(case_error)
    return errors


def validate_volume_interval(order: OrderWithCases, **kwargs) -> list[InvalidVolumeError]:
    errors: list[InvalidVolumeError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_volume_invalid(sample):
                error = InvalidVolumeError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_volume_required(order: OrderWithCases, **kwargs) -> list[VolumeRequiredError]:
    errors: list[VolumeRequiredError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_volume_missing(sample):
                error = VolumeRequiredError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_samples_exist(
    order: OrderWithCases,
    store: Store,
    **kwargs,
) -> list[SampleDoesNotExistError]:
    errors: list[SampleDoesNotExistError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_existing_samples:
            sample: DbSample | None = store.get_sample_by_internal_id(sample.internal_id)
            if not sample:
                error = SampleDoesNotExistError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_wells_contain_at_most_one_sample(
    order: OrderWithCases, **kwargs
) -> list[OccupiedWellError]:
    errors: list[OccupiedWellError] = []
    well_position_to_sample_map: dict[tuple[str, str], list[tuple[int, int]]] = get_well_sample_map(
        order
    )
    for indices in well_position_to_sample_map.values():
        if len(indices) > 1:
            well_errors = get_occupied_well_errors(indices[1:])
            errors.extend(well_errors)
    return errors


def validate_sample_names_not_repeated(
    order: OrderWithCases, store: Store, **kwargs
) -> list[SampleNameRepeatedError]:
    """Ensures that sample names are unique within the order
    and that they not already used in the case previously."""
    old_sample_names: set[str] = get_existing_sample_names(order=order, status_db=store)
    new_samples: list[tuple[int, int, SampleInCase]] = order.enumerated_new_samples
    sample_name_counter = Counter([sample.name for _, _, sample in new_samples])
    return [
        SampleNameRepeatedError(case_index=case_index, sample_index=sample_index)
        for case_index, sample_index, sample in new_samples
        if sample_name_counter.get(sample.name) > 1 or sample.name in old_sample_names
    ]


def validate_sample_names_different_from_case_names(
    order: OrderWithCases, store: Store, **kwargs
) -> list[SampleNameSameAsCaseNameError]:
    """Return errors with the indexes of samples having the same name as any case in the order."""
    errors: list[SampleNameSameAsCaseNameError] = []
    new_case_names: set[str] = {case.name for _, case in order.enumerated_new_cases}
    existing_case_names: set[str] = get_existing_case_names(order=order, status_db=store)
    all_case_names = new_case_names.union(existing_case_names)
    for case_index, sample_index, sample in order.enumerated_new_samples:
        if sample.name in all_case_names:
            error = SampleNameSameAsCaseNameError(
                case_index=case_index,
                sample_index=sample_index,
            )
            errors.append(error)
    return errors


def validate_fathers_are_male(
    order: OrderWithCases, store: Store, **kwargs
) -> list[InvalidFatherSexError]:
    errors: list[InvalidFatherSexError] = []
    for index, case in order.enumerated_new_cases:
        case_errors: list[InvalidFatherSexError] = get_father_sex_errors(
            case=case, case_index=index, store=store
        )
        errors.extend(case_errors)
    return errors


def validate_fathers_in_same_case_as_children(
    order: OrderWithCases, store: Store, **kwargs
) -> list[FatherNotInCaseError]:
    errors: list[FatherNotInCaseError] = []
    for index, case in order.enumerated_new_cases:
        case_errors: list[FatherNotInCaseError] = get_father_case_errors(
            case=case, case_index=index, store=store
        )
        errors.extend(case_errors)
    return errors


def validate_mothers_are_female(
    order: OrderWithCases, store: Store, **kwargs
) -> list[InvalidMotherSexError]:
    errors: list[InvalidMotherSexError] = []
    for index, case in order.enumerated_new_cases:
        case_errors: list[InvalidMotherSexError] = get_mother_sex_errors(
            case=case, case_index=index, store=store
        )
        errors.extend(case_errors)
    return errors


def validate_mothers_in_same_case_as_children(
    order: OrderWithCases, store: Store, **kwargs
) -> list[MotherNotInCaseError]:
    errors: list[MotherNotInCaseError] = []
    for index, case in order.enumerated_new_cases:
        case_errors: list[MotherNotInCaseError] = get_mother_case_errors(
            case=case, case_index=index, store=store
        )
        errors.extend(case_errors)
    return errors


def validate_pedigree(order: OrderWithCases, store: Store, **kwargs) -> list[PedigreeError]:
    errors: list[PedigreeError] = []
    for case_index, case in order.enumerated_new_cases:
        case_errors: list[PedigreeError] = get_pedigree_errors(
            case=case, case_index=case_index, store=store
        )
        errors.extend(case_errors)
    return errors


def validate_subject_sex_consistency(
    order: OrderWithCases,
    store: Store,
) -> list[SexSubjectIdError]:
    errors: list[SexSubjectIdError] = []

    for case_index, sample_index, sample in order.enumerated_new_samples:
        if not has_sex_and_subject(sample):
            continue
        if store.sample_exists_with_different_sex(
            customer_internal_id=order.customer,
            subject_id=sample.subject_id,
            sex=sample.sex,
        ):
            error = SexSubjectIdError(
                case_index=case_index,
                sample_index=sample_index,
            )
            errors.append(error)
    return errors


def validate_subject_ids_different_from_case_names(
    order: OrderWithCases, **kwargs
) -> list[SubjectIdSameAsCaseNameError]:
    errors: list[SubjectIdSameAsCaseNameError] = []
    for index, case in order.enumerated_new_cases:
        case_errors: list[SubjectIdSameAsCaseNameError] = validate_subject_ids_in_case(
            case=case,
            case_index=index,
        )
        errors.extend(case_errors)
    return errors


def validate_concentration_interval_if_skip_rc(
    order: OrderWithCases, store: Store, **kwargs
) -> list[InvalidConcentrationIfSkipRCError]:
    if not order.skip_reception_control:
        return []
    errors: list[InvalidConcentrationIfSkipRCError] = []
    for index, case in order.enumerated_new_cases:
        case_errors: list[InvalidConcentrationIfSkipRCError] = validate_concentration_in_case(
            case=case,
            case_index=index,
            store=store,
        )
        errors.extend(case_errors)
    return errors


def validate_well_position_format(order: OrderWithCases, **kwargs) -> list[WellFormatError]:
    errors: list[WellFormatError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_invalid_plate_well_format(sample=sample):
                error = WellFormatError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_tube_container_name_unique(
    order: OrderWithCases, **kwargs
) -> list[ContainerNameRepeatedError]:
    errors: list[ContainerNameRepeatedError] = []

    container_name_counter: Counter = get_counter_container_names(order)

    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_sample_tube_name_reused(sample=sample, counter=container_name_counter):
                error = ContainerNameRepeatedError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_not_all_samples_unknown_in_case(
    order: OrderWithCases, **kwargs
) -> list[StatusUnknownError]:
    errors: list[StatusUnknownError] = []

    for case_index, case in order.enumerated_new_cases:
        if are_all_samples_unknown(case):
            for sample_index, _ in case.enumerated_samples:
                error = StatusUnknownError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_buffer_required(order: OrderWithCases, **kwargs) -> list[BufferMissingError]:
    """Return an error for each new sample missing a buffer, if its application requires one."""

    errors: list[BufferMissingError] = []
    for case_index, sample_index, sample in order.enumerated_new_samples:
        if is_buffer_missing(sample):
            error = BufferMissingError(case_index=case_index, sample_index=sample_index)
            errors.append(error)
    return errors


def reset_optional_capture_kits(
    order: BalsamicOrder | BalsamicUmiOrder, store: Store, **kwargs
) -> list[CaptureKitResetError]:
    """
    Sets the capture kit to None for each sample where it is set but not needed and returns an error
    to be rendered as a warning for each such sample.
    """
    errors: list[CaptureKitResetError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if not does_sample_need_capture_kit(sample=sample, store=store) and sample.capture_kit:
                sample.capture_kit = None
                error = CaptureKitResetError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_capture_kit_requirement(
    order: BalsamicOrder | BalsamicUmiOrder, store: Store
) -> list[CaptureKitMissingError]:
    """
    Return an error for each new sample missing a capture kit, if its application requires one.
    Applicable to Balsamic and Balsamic-UMI orders only.
    """
    errors: list[CaptureKitMissingError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_new_samples:
            if is_sample_missing_capture_kit(sample=sample, store=store):
                error = CaptureKitMissingError(case_index=case_index, sample_index=sample_index)
                errors.append(error)
    return errors


def validate_capture_kit(
    order: BalsamicOrder | BalsamicUmiOrder, store: Store
) -> list[InvalidCaptureKitError]:
    errors: list[InvalidCaptureKitError] = []
    for case_index, sample_index, sample in order.enumerated_new_samples:
        if is_invalid_capture_kit(sample=sample, store=store):
            errors.append(InvalidCaptureKitError(sample_index=sample_index, case_index=case_index))

    return errors


def validate_existing_samples_belong_to_collaboration(
    order: OrderWithCases, store: Store, **kwargs
) -> list[SampleOutsideOfCollaborationError]:
    """Validates that existing samples belong to the same collaboration as the order's customer."""
    errors: list[SampleOutsideOfCollaborationError] = []
    for case_index, case in order.enumerated_new_cases:
        for sample_index, sample in case.enumerated_existing_samples:
            if is_sample_not_from_collaboration(
                customer_id=order.customer, sample=sample, store=store
            ):
                error = SampleOutsideOfCollaborationError(
                    sample_index=sample_index, case_index=case_index
                )
                errors.append(error)
    return errors


def validate_existing_samples_compatible_with_order_type(
    order: OrderWithCases, store: Store, **kwargs
) -> list[ExistingSampleWrongTypeError]:
    errors: list[ExistingSampleWrongTypeError] = []
    for case_index, sample_index, sample in order.enumerated_existing_samples:
        if not is_sample_compatible_with_order_type(
            order_type=order.order_type, sample=sample, store=store
        ):
            error = ExistingSampleWrongTypeError(case_index=case_index, sample_index=sample_index)
            errors.append(error)
    return errors


def validate_sample_names_available(
    order: OrderWithCases, store: Store, **kwargs
) -> list[SampleNameAlreadyExistsError]:
    """Validates that new sample names are not already used by the customer."""
    errors: list[SampleNameAlreadyExistsError] = []
    customer_entry_id: int = store.get_customer_by_internal_id(order.customer).id
    for case_index, sample_index, sample in order.enumerated_new_samples:
        if store.is_sample_name_used(sample=sample, customer_entry_id=customer_entry_id):
            error = SampleNameAlreadyExistsError(case_index=case_index, sample_index=sample_index)
            errors.append(error)
    return errors



"""Loqusdb related constants."""

from enum import Enum, StrEnum

from cg.constants.constants import CancerAnalysisType, CustomerId, Workflow
from cg.constants.sequencing import SeqLibraryPrepCategory

LOQUSDB_ID = "_id"
LOQUSDB_SUPPORTED_WORKFLOWS = [
    Workflow.BALSAMIC,
    Workflow.MIP_DNA,
    Workflow.NALLO,
    Workflow.RAREDISEASE,
]
LOQUSDB_RARE_DISEASE_CUSTOMERS = [CustomerId.CUST002, CustomerId.CUST003, CustomerId.CUST004]
LOQUSDB_CANCER_CUSTOMERS = [
    CustomerId.CUST110,
    CustomerId.CUST127,
    CustomerId.CUST143,
    CustomerId.CUST147,
]
LOQUSDB_LONG_READ_CUSTOMERS = [
    CustomerId.CUST002,
    CustomerId.CUST003,
    CustomerId.CUST004,
    CustomerId.CUST201,
]
LOQUSDB_LONG_READ_SEQUENCING_METHODS = [
    SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
]
LOQUSDB_RARE_DISEASE_SEQUENCING_METHODS = [
    SeqLibraryPrepCategory.WHOLE_GENOME_SEQUENCING,
    SeqLibraryPrepCategory.WHOLE_EXOME_SEQUENCING,
]
LOQUSDB_CANCER_SEQUENCING_METHODS = [
    CancerAnalysisType.TUMOR_WGS,
    CancerAnalysisType.TUMOR_NORMAL_WGS,
]


class BalsamicLoadParameters(Enum):
    """Cancer Loqusdb load command parameters."""

    QUAL_THRESHOLD: int = 0
    QUAL_GERMLINE_THRESHOLD: int = 10


class BalsamicObservationsAnalysisTag(StrEnum):
    """Cancer observations files analysis tags."""

    SNV_GERMLINE_VCF: str = "vcf-snv-germline-tumor"
    SNV_VCF: str = "vcf-snv-clinical"
    SV_GERMLINE_VCF: str = "vcf-sv-germline-tumor"
    SV_VCF: str = "vcf-sv-clinical"


class LoqusdbInstance(StrEnum):
    """Observations instances."""

    LWP: str = "loqusdb-lwp"
    WGS: str = "loqusdb"
    WES: str = "loqusdb-wes"
    SOMATIC: str = "loqusdb-somatic"
    TUMOR: str = "loqusdb-tumor"


class MipDNALoadParameters(Enum):
    """Rare disease Loqusdb load command parameters."""

    PROFILE_THRESHOLD: float = 0.95
    GQ_THRESHOLD: int = 10
    HARD_THRESHOLD: float = 0.95
    SOFT_THRESHOLD: float = 0.90


class MipDNAObservationsAnalysisTag(StrEnum):
    """Rare disease observations files analysis tags."""

    SNV_VCF: str = "deepvariant"
    SV_VCF: str = "vcf-sv-research"
    PROFILE_GBCF: str = "snv-gbcf"
    FAMILY_PED: str = "pedigree"


class NalloObservationsAnalysisTag(StrEnum):
    """Nallo observations files analysis tags."""

    SNV_VCF = "vcf-snv-research"
    SV_VCF = "vcf-sv-research"
    FAMILY_PED = "pedigree"


class NalloLoadParameters(Enum):
    """Nallo Loqusdb load command parameters."""

    PROFILE_THRESHOLD = 0.95
    GQ_THRESHOLD = 10
    HARD_THRESHOLD = 0.95
    SOFT_THRESHOLD = 0.90
    SNV_GQ_ONLY = True


class ObservationsFileWildcards(StrEnum):
    """File patterns regarding dump Loqusdb files."""

    ARTEFACT_SNV: str = "artefact_somatic_snv"
    CLINICAL_SNV: str = "clinical_snv"
    CLINICAL_SV: str = "clinical_sv"
    CANCER_GERMLINE_SNV: str = "cancer_germline_snv"
    CANCER_GERMLINE_SV: str = "cancer_germline_sv"
    CANCER_SOMATIC_SNV: str = "cancer_somatic_snv"
    CANCER_SOMATIC_SV: str = "cancer_somatic_sv"


class RarediseaseObservationsAnalysisTag(StrEnum):
    """Rare disease observations files analysis tags."""

    SNV_VCF: str = "vcf-snv"
    SV_VCF: str = "vcf-sv"
    FAMILY_PED: str = "pedigree"


class RarediseaseLoadParameters(Enum):
    """Raredisease Loqusdb load command parameters."""

    PROFILE_THRESHOLD: float = 0.95
    GQ_THRESHOLD: int = 10
    HARD_THRESHOLD: float = 0.95
    SOFT_THRESHOLD: float = 0.90



from enum import StrEnum

from cg.constants import DataDelivery


class MicrosaltDeliveryType(StrEnum):
    FASTQ_QC = DataDelivery.FASTQ_QC
    FASTQ_QC_ANALYSIS = DataDelivery.FASTQ_QC_ANALYSIS
    NO_DELIVERY = DataDelivery.NO_DELIVERY



from pydantic import Discriminator, Tag
from typing_extensions import Annotated

from cg.services.orders.validation.models.discriminators import has_internal_id
from cg.services.orders.validation.models.existing_case import ExistingCase
from cg.services.orders.validation.models.order_with_cases import OrderWithCases
from cg.services.orders.validation.order_types.mip_rna.constants import MIPRNADeliveryType
from cg.services.orders.validation.order_types.mip_rna.models.case import MIPRNACase

NewCase = Annotated[MIPRNACase, Tag("new")]
OldCase = Annotated[ExistingCase, Tag("existing")]


class MIPRNAOrder(OrderWithCases):
    cases: list[Annotated[NewCase | OldCase, Discriminator(has_internal_id)]]
    delivery_type: MIPRNADeliveryType



from cg.constants.constants import Strandedness
from cg.models.nf_analysis import NextflowSampleSheetEntry, WorkflowParameters
from cg.models.qc_metrics import QCMetrics


class RnafusionQCMetrics(QCMetrics):
    """RNAfusion QC metrics."""

    after_filtering_gc_content: float
    after_filtering_q20_rate: float
    after_filtering_q30_rate: float
    after_filtering_read1_mean_length: float
    before_filtering_total_reads: float
    median_5prime_to_3prime_bias: float
    pct_adapter: float
    pct_mrna_bases: float
    pct_ribosomal_bases: float
    pct_surviving: float
    pct_duplication: float
    read_pairs_examined: float
    uniquely_mapped_percent: float


class RnafusionParameters(WorkflowParameters):
    """Rnafusion parameters."""


class RnafusionSampleSheetEntry(NextflowSampleSheetEntry):
    """Rnafusion sample sheet model."""

    strandedness: Strandedness

    @staticmethod
    def headers() -> list[str]:
        """Return sample sheet headers."""
        return ["sample", "fastq_1", "fastq_2", "strandedness"]

    def reformat_sample_content(self) -> list[list[str]]:
        """Reformat sample sheet content as a list of list, where each list represents a line in the final file."""
        return [
            [self.name, fastq_forward_read_path, fastq_reverse_read_path, str(self.strandedness)]
            for fastq_forward_read_path, fastq_reverse_read_path in zip(
                self.fastq_forward_read_paths, self.fastq_reverse_read_paths
            )
        ]



from pydantic import BaseModel, Field, field_validator

from cg.constants import Workflow
from cg.server.dto.orders.orders_request import OrderSortField, SortOrder


class OrderQueryParams(BaseModel):
    page_size: int | None = Field(default=50)
    page: int | None = Field(default=1)
    sort_field: str | None = Field(default=OrderSortField.ORDER_DATE)
    sort_order: str | None = Field(default=SortOrder.ASC)
    search: str | None = None
    workflows: list[str] | None = []
    is_open: bool | None = None

    @field_validator("workflows", mode="before")
    def expand_balsamic_workflow(cls, value):
        """Expand the BALSAMIC workflow to include the UMI and QC workflows when Trailblazer request Balsamic."""
        if value and Workflow.BALSAMIC in value:
            value = [Workflow.BALSAMIC, Workflow.BALSAMIC_UMI]
        return value



from enum import IntEnum, StrEnum


class Sex(StrEnum):
    FEMALE = "female"
    MALE = "male"
    UNKNOWN = "unknown"
    OTHER = "other"
    MISSING = ""


class PhenotypeStatus(StrEnum):
    UNKNOWN = "unknown"
    UNAFFECTED = "unaffected"
    AFFECTED = "affected"
    MISSING = ""


class PlinkPhenotypeStatus(IntEnum):
    UNKNOWN = 0
    UNAFFECTED = 1
    AFFECTED = 2


class PlinkSex(StrEnum):
    UNKNOWN = str(0)
    MALE = str(1)
    FEMALE = str(2)


class RelationshipStatus(StrEnum):
    HAS_NO_PARENT = str(0)


SCOUT_PRIORITIZED_STATUS: str = "prioritized"



import rich_click as click

from cg.constants.cli_options import DRY_RUN
from cg.services.illumina.post_processing.housekeeper_storage import (
    delete_sequencing_data_from_housekeeper,
)


@click.command("illumina-run")
@click.argument(
    "flow_cell_id",
    type=str,
)
@click.pass_context
def delete_illumina_run(context: click.Context, flow_cell_id: str):
    """
    Delete an Illumina sequencing run.

    Args:
        flow_cell_id (str): Give the flow cell id of the run to delete from statusdb and housekeeper.
    """
    delete_sequencing_data_from_housekeeper(
        flow_cell_id=flow_cell_id,
        hk_api=context.obj.housekeeper_api,
    )
    context.obj.status_db.delete_illumina_flow_cell(internal_id=flow_cell_id)



import logging
import shutil
from datetime import datetime
from pathlib import Path

import rich_click as click

from cg.constants import EXIT_FAIL, EXIT_SUCCESS, Workflow
from cg.constants.constants import FileExtensions
from cg.constants.tb import AnalysisStatus
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import MicrosaltFastqHandler
from cg.meta.workflow.microsalt.quality_controller import MicroSALTQualityController
from cg.meta.workflow.microsalt.quality_controller.models import QualityResult
from cg.meta.workflow.microsalt.utils import get_most_recent_project_directory
from cg.models.cg_config import CGConfig
from cg.store.models import Case, Sample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class MicrosaltAnalysisAPI(AnalysisAPI):
    """API to manage Microsalt Analyses"""

    def __init__(self, config: CGConfig, workflow: Workflow = Workflow.MICROSALT):
        super().__init__(workflow, config)
        self.root_dir = config.microsalt.root
        self.queries_path = config.microsalt.queries_path
        self.quality_checker = MicroSALTQualityController(config.status_db)

    @property
    def fastq_handler(self):
        return MicrosaltFastqHandler

    @property
    def conda_binary(self) -> str:
        return self.config.microsalt.conda_binary

    @property
    def process(self) -> Process:
        if not self._process:
            self._process = Process(
                binary=self.config.microsalt.binary_path,
                conda_binary=f"{self.conda_binary}" if self.conda_binary else None,
                environment=self.config.microsalt.conda_env,
            )
        return self._process

    def clean_run_dir(
        self, case_id: str, skip_confirmation: bool, case_path: list[Path] | Path
    ) -> int:
        """Remove workflow run directories for a MicroSALT case."""

        if not case_path:
            LOG.info(
                f"There is no case paths for case {case_id}. Setting cleaned at to {datetime.now()}"
            )
            self.clean_analyses(case_id=case_id)
            return EXIT_SUCCESS
        if isinstance(case_path, Path):
            case_path: list[Path] = [case_path]
        for analysis_path in case_path:
            if skip_confirmation or click.confirm(
                f"Are you sure you want to remove all files in {analysis_path}?"
            ):
                if analysis_path.is_symlink():
                    LOG.warning(
                        f"Will not automatically delete symlink: {analysis_path}, delete it manually",
                    )
                    return EXIT_FAIL

                shutil.rmtree(analysis_path, ignore_errors=True)
                LOG.info(f"Cleaned {analysis_path}")

        self.clean_analyses(case_id=case_id)
        return EXIT_SUCCESS

    def get_case_fastq_path(self, case_id: str) -> Path:
        """Get fastq paths for a case."""
        return Path(self.root_dir, "fastq", case_id)

    def get_config_path(self, filename: str) -> Path:
        return Path(self.queries_path, filename).with_suffix(".json")

    def get_job_ids_path(self, case_id: str) -> Path:
        project_id: str = self.get_project_id(case_id)
        return Path(
            self.root_dir,
            "results",
            "reports",
            "trailblazer",
            f"{project_id}_slurm_ids{FileExtensions.YAML}",
        )

    def get_deliverables_file_path(self, case_id: str) -> Path:
        """Returns a path where the microSALT deliverables file for the order_id should be
        located"""
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        order_id: str = case_obj.name
        deliverables_file_path = Path(
            self.root_dir,
            "results",
            "reports",
            "deliverables",
            f"{order_id}_deliverables.yaml",
        )
        if deliverables_file_path.exists():
            LOG.info("Found deliverables file %s", deliverables_file_path)
        return deliverables_file_path

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        return Path(self.get_case_fastq_path(case_id=case.internal_id), sample.internal_id)

    def link_fastq_files(self, case_id: str, sample_id: str | None, dry_run: bool = False) -> None:
        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        samples: list[Sample] = self.get_samples(case_id=case_id, sample_id=sample_id)
        for sample_obj in samples:
            self.link_fastq_files_for_sample(case=case_obj, sample=sample_obj)

    def get_samples(self, case_id: str, sample_id: str | None = None) -> list[Sample]:
        """Returns a list of samples to configure
        If sample_id is specified, will return a list with only this sample_id.
        Otherwise, returns all samples in given case"""
        if sample_id:
            return [self.status_db.get_sample_by_internal_id(internal_id=sample_id)]

        case_obj: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        return [link.sample for link in case_obj.links]

    def get_cases_to_store(self) -> list[Case]:
        cases_qc_ready: list[Case] = self.get_completed_cases()
        cases_to_store: list[Case] = []
        LOG.info(f"Found {len(cases_qc_ready)} cases to perform QC on!")

        for case in cases_qc_ready:
            case_run_dir: Path | None = self.get_case_path(case.internal_id)
            LOG.info(f"Checking QC for case {case.internal_id} in {case_run_dir}")

            if self.quality_checker.is_qc_required(case_run_dir):
                LOG.info(f"QC required for case {case.internal_id}")
                metrics_file_path = self.get_metrics_file_path(case.internal_id)

                if not metrics_file_path.exists():
                    continue

                result: QualityResult = self.quality_checker.quality_control(metrics_file_path)
                self.trailblazer_api.add_comment(case_id=case.internal_id, comment=result.summary)
                if result.passes_qc:
                    cases_to_store.append(case)
                else:
                    self.trailblazer_api.set_analysis_status(
                        case_id=case.internal_id, status=AnalysisStatus.FAILED
                    )
            else:
                cases_to_store.append(case)

        return cases_to_store

    def get_completed_cases(self) -> list[Case]:
        """Return cases that are completed in trailblazer."""
        return [
            case
            for case in self.status_db.get_running_cases_in_workflow(self.workflow)
            if self.trailblazer_api.is_latest_analysis_completed(case.internal_id)
        ]

    def get_metrics_file_path(self, case_id: str) -> Path:
        """Return path to metrics file for a case."""
        project_id: str = self.get_project_id(case_id)
        case_run_dir: Path = self.get_case_path(case_id)
        return Path(case_run_dir, f"{project_id}{FileExtensions.JSON}")

    def extract_project_id(self, sample_id: str) -> str:
        return sample_id.rsplit("A", maxsplit=1)[0]

    def get_project_id(self, case_id: str) -> str:
        case: Case = self.status_db.get_case_by_internal_id(case_id)
        sample_id: str = case.links[0].sample.internal_id
        return self.extract_project_id(sample_id)

    def get_results_dir(self) -> Path:
        return Path(self.root_dir, "results")

    def get_case_path(self, case_id: str) -> Path:
        project_id: str = self.get_project_id(case_id)
        results_dir: Path = self.get_results_dir()
        return get_most_recent_project_directory(project_id=project_id, directory=results_dir)



"""Module to handle time functions"""

import re
import time
from datetime import datetime
from pathlib import Path

from cg.utils.files import get_source_creation_time_stamp

DATE_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S"


def get_start_time() -> float:
    """Returns the current time"""
    return time.time()


def format_time_from_string(time_stamp: str) -> datetime:
    """Format the time by removing microseconds and time zone."""
    no_microseconds: str = re.sub(r"\.\d+", "", time_stamp)
    no_microseconds_no_time_zone = re.sub(r"[+-]\d{2}:\d{2}$", "", no_microseconds)
    return datetime.strptime(no_microseconds_no_time_zone, DATE_TIME_FORMAT)


def format_time_from_ctime(time_stamp: float) -> datetime:
    """Format the time by removing microseconds and time zone."""
    dt: datetime = datetime.fromtimestamp(
        time_stamp,
        tz=None,
    )
    dt_without_microseconds: datetime = dt.replace(microsecond=0)
    return dt_without_microseconds


def get_elapsed_time(start_time: float) -> float:
    """Determines the time elapsed since the start time"""
    return time.time() - start_time


def is_directory_older_than_days_old(directory_path: Path, days_old: int) -> bool:
    """
    Check if the directory is older than the specified number of days.
    """
    days_in_seconds = 60 * 60 * 24 * days_old
    dir_creation_time_stamp: float = get_source_creation_time_stamp(source_path=directory_path)
    return bool(dir_creation_time_stamp < time.time() - days_in_seconds)



"""Store class that combines all CRUD operations for the CG database."""

from cg.store.crud.create import CreateMixin
from cg.store.crud.delete import DeleteMixin
from cg.store.crud.update import UpdateMixin


class Store(
    CreateMixin,
    DeleteMixin,
    UpdateMixin,
):
    pass



import logging
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.io.csv import write_csv
from cg.store.models import Case, CaseSample, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)

HEADERS: list[str] = [
    "project",
    "sample",
    "file",
    "family_id",
    "paternal_id",
    "maternal_id",
    "sex",
    "phenotype",
]


class NalloSampleSheetCreator:
    def __init__(self, housekeeper_api: HousekeeperAPI, status_db: Store) -> None:
        self.housekeeper_api = housekeeper_api
        self.status_db = status_db

    def create(self, case_id: str, file_path: Path) -> None:
        LOG.debug(f"Creating sample sheet for case {case_id}")
        content: list[list[str]] = self._get_content(case_id)
        write_csv(content=content, file_path=file_path)

    def _get_content(self, case_id: str) -> list[list[str]]:
        case: Case = self.status_db.get_case_by_internal_id_strict(case_id)
        sample_sheet_content: list[list[str]] = [HEADERS]
        for link in case.links:
            sample_sheet_content.extend(self._get_sample_sheet_content_per_sample(case_sample=link))
        return sample_sheet_content

    def _get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        read_file_paths: list[str] = self._get_bam_read_file_paths(sample=case_sample.sample)
        sample_sheet_entries = []

        for bam_path in read_file_paths:
            sample_sheet_entry: list[str] = [
                case_sample.case.internal_id,
                case_sample.sample.internal_id,
                bam_path,
                case_sample.case.internal_id,
                case_sample.get_paternal_sample_id or "0",
                case_sample.get_maternal_sample_id or "0",
                PlinkSex[case_sample.sample.sex.upper()].value,
                str(PlinkPhenotypeStatus[case_sample.status.upper()]),
            ]
            sample_sheet_entries.append(sample_sheet_entry)
        return sample_sheet_entries

    def _get_bam_read_file_paths(self, sample: Sample) -> list[str]:
        """Gather BAM file path for a sample based on the BAM tag."""
        return [
            hk_file.full_path
            for hk_file in self.housekeeper_api.files(bundle=sample.internal_id, tags={"bam"})
        ]



from cg.constants.constants import DataDelivery, MicrosaltAppTags, Workflow
from cg.exc import CaseNotFoundError, OrderMismatchError
from cg.services.delivery_message.messages import (
    AnalysisScoutMessage,
    CovidMessage,
    FastqAnalysisScoutMessage,
    FastqMessage,
    FastqScoutMessage,
    MicrosaltMwrMessage,
    ScoutMessage,
    StatinaMessage,
)
from cg.services.delivery_message.messages.analysis_message import AnalysisMessage
from cg.services.delivery_message.messages.bam_message import BamMessage
from cg.services.delivery_message.messages.delivery_message import DeliveryMessage
from cg.services.delivery_message.messages.fastq_analysis_message import FastqAnalysisMessage
from cg.services.delivery_message.messages.microsalt_mwx_message import MicrosaltMwxMessage
from cg.services.delivery_message.messages.order_message import TaxprofilerDeliveryMessage
from cg.services.delivery_message.messages.raw_data_analysis_message import RawDataAnalysisMessage
from cg.services.delivery_message.messages.raw_data_analysis_scout_message import (
    RawDataAnalysisScoutMessage,
)
from cg.services.delivery_message.messages.raw_data_scout_message import RawDataScoutMessage
from cg.services.delivery_message.messages.rna_delivery_message import (
    RNAAnalysisStrategy,
    RNADeliveryMessage,
    RNAFastqAnalysisStrategy,
    RNAFastqStrategy,
    RNAScoutStrategy,
    RNAUploadMessageStrategy,
)
from cg.store.models import Case, Sample
from cg.store.store import Store

MESSAGE_MAP = {
    DataDelivery.ANALYSIS_FILES: AnalysisMessage,
    DataDelivery.ANALYSIS_SCOUT: AnalysisScoutMessage,
    DataDelivery.BAM: BamMessage,
    DataDelivery.FASTQ: FastqMessage,
    DataDelivery.FASTQ_ANALYSIS: FastqAnalysisMessage,
    DataDelivery.FASTQ_ANALYSIS_SCOUT: FastqAnalysisScoutMessage,
    DataDelivery.FASTQ_SCOUT: FastqScoutMessage,
    DataDelivery.RAW_DATA_ANALYSIS: RawDataAnalysisMessage,
    DataDelivery.RAW_DATA_ANALYSIS_SCOUT: RawDataAnalysisScoutMessage,
    DataDelivery.RAW_DATA_SCOUT: RawDataScoutMessage,
    DataDelivery.SCOUT: ScoutMessage,
    DataDelivery.STATINA: StatinaMessage,
}


RNA_STRATEGY_MAP: dict[DataDelivery, type[RNAUploadMessageStrategy]] = {
    # Only returns a message strategy if there is a scout delivery for the case.
    DataDelivery.SCOUT: RNAScoutStrategy,
    DataDelivery.FASTQ_SCOUT: RNAFastqStrategy,
    DataDelivery.ANALYSIS_SCOUT: RNAAnalysisStrategy,
    DataDelivery.FASTQ_ANALYSIS_SCOUT: RNAFastqAnalysisStrategy,
}


def get_message(cases: list[Case], store: Store) -> str:
    message_strategy: DeliveryMessage = get_message_strategy(case=cases[0], store=store)
    return message_strategy.create_message(cases)


def get_message_strategy(case: Case, store: Store) -> DeliveryMessage:
    if case.data_analysis == Workflow.MICROSALT:
        return get_microsalt_message_strategy(case)

    if case.data_analysis == Workflow.MUTANT:
        return CovidMessage()

    if case.data_analysis in [Workflow.MIP_RNA, Workflow.TOMTE]:
        return get_rna_message_strategy_from_data_delivery(case=case, store=store)

    if case.data_analysis in Workflow.TAXPROFILER:
        return TaxprofilerDeliveryMessage()

    message_strategy: DeliveryMessage = get_message_strategy_from_data_delivery(case)
    return message_strategy


def get_message_strategy_from_data_delivery(case: Case) -> DeliveryMessage:
    message_strategy: DeliveryMessage = MESSAGE_MAP[case.data_delivery]()
    return message_strategy


def get_rna_message_strategy_from_data_delivery(
    case: Case, store: Store
) -> DeliveryMessage | RNADeliveryMessage:
    """Get the RNA delivery message strategy based on the data delivery type.
    If a scout delivery is required it will use the RNADeliveryMessage class that links RNA to DNA cases.
    Otherwise it used the conventional delivery message strategy.
    """
    if message_strategy := RNA_STRATEGY_MAP.get(case.data_delivery):
        return RNADeliveryMessage(store=store, strategy=message_strategy())
    return MESSAGE_MAP[case.data_delivery]()


def get_microsalt_message_strategy(case: Case) -> DeliveryMessage:
    if has_mwx_samples(case) or has_vwg_samples(case):
        return MicrosaltMwxMessage()

    if has_mwr_samples(case):
        return MicrosaltMwrMessage()

    app_tag: str = get_case_app_tag(case)
    raise NotImplementedError(f"Microsalt apptag {app_tag} not supported.")


def has_mwx_samples(case: Case) -> bool:
    case_app_tag: str = get_case_app_tag(case)
    return case_app_tag == MicrosaltAppTags.MWXNXTR003


def has_mwr_samples(case: Case) -> bool:
    case_app_tag: str = get_case_app_tag(case)
    return case_app_tag == MicrosaltAppTags.MWRNXTR003


def has_vwg_samples(case: Case) -> bool:
    case_app_tag: str = get_case_app_tag(case)
    return case_app_tag == MicrosaltAppTags.VWGNXTR001


def get_case_app_tag(case: Case) -> str:
    sample: Sample = case.samples[0]
    return get_sample_app_tag(sample)


def get_sample_app_tag(sample: Sample) -> str:
    return sample.application_version.application.tag


def validate_cases(cases: list[Case], case_ids: list[str]) -> None:
    if set(case_ids) != set(case.internal_id for case in cases):
        raise CaseNotFoundError("Internal id not found in the database")
    if not is_matching_order(cases):
        raise OrderMismatchError("Cases do not belong to the same order")


def is_matching_order(cases: list[Case]) -> bool:
    return all([case.latest_order == cases[0].latest_order for case in cases])



from pathlib import Path

from pydantic import BaseModel


class DeliveryMetaData(BaseModel):
    case_id: str
    customer_internal_id: str
    ticket_id: str
    delivery_path: Path | None = None


class CaseFile(BaseModel):
    case_id: str
    case_name: str
    file_path: Path


class SampleFile(BaseModel):
    case_id: str
    sample_id: str
    sample_name: str
    file_path: Path


class DeliveryFiles(BaseModel):
    delivery_data: DeliveryMetaData
    case_files: list[CaseFile]
    sample_files: list[SampleFile]



import logging
from pathlib import Path

from cg.services.deliver_files.file_fetcher.models import CaseFile, DeliveryFiles, SampleFile
from cg.services.deliver_files.file_formatter.destination.abstract import (
    DeliveryDestinationFormatter,
)
from cg.services.deliver_files.file_formatter.destination.models import (
    FormattedFile,
    FormattedFiles,
)
from cg.services.deliver_files.file_formatter.files.case_service import CaseFileFormatter
from cg.services.deliver_files.file_formatter.files.mutant_service import (
    MutantFileFormatter,
)
from cg.services.deliver_files.file_formatter.files.concatenation_service import (
    SampleFileConcatenationFormatter,
)
from cg.services.deliver_files.file_formatter.files.sample_service import (
    SampleFileFormatter,
)

LOG = logging.getLogger(__name__)


class BaseDeliveryFormatter(DeliveryDestinationFormatter):
    """
    Format the files to be delivered in the generic format.
    args:
        case_file_formatter: The case file formatter
        sample_file_formatter: The sample file formatter. This can be a SampleFileFormatter, SampleFileConcatenationFormatter or MutantFileFormatter.
    """

    def __init__(
        self,
        case_file_formatter: CaseFileFormatter,
        sample_file_formatter: (
            SampleFileFormatter | SampleFileConcatenationFormatter | MutantFileFormatter
        ),
    ):
        self.case_file_formatter = case_file_formatter
        self.sample_file_formatter = sample_file_formatter

    def format_files(self, delivery_files: DeliveryFiles) -> FormattedFiles:
        """Format the files to be delivered and return the formatted files in the generic format."""
        LOG.debug("[FORMAT SERVICE] Formatting files for delivery")
        formatted_files: list[FormattedFile] = self._format_sample_and_case_files(
            sample_files=delivery_files.sample_files,
            case_files=delivery_files.case_files,
            delivery_path=delivery_files.delivery_data.delivery_path,
        )
        return FormattedFiles(files=formatted_files)

    def _format_sample_and_case_files(
        self, sample_files: list[SampleFile], case_files: list[CaseFile], delivery_path: Path
    ) -> list[FormattedFile]:
        """Helper method to format both sample and case files."""
        LOG.debug(f"[FORMAT SERVICE] delivery_path: {delivery_path}")
        formatted_files: list[FormattedFile] = self.sample_file_formatter.format_files(
            moved_files=sample_files,
            delivery_path=delivery_path,
        )
        if case_files:
            formatted_case_files: list[FormattedFile] = self.case_file_formatter.format_files(
                moved_files=case_files,
                delivery_path=delivery_path,
            )
            formatted_files.extend(formatted_case_files)
        return formatted_files



from pathlib import Path

from cg.constants.priority import SlurmQos
from cg.io.yaml import read_yaml, write_yaml_stream
from cg.services.analysis_starter.configurator.models.nextflow import NextflowCaseConfig
from cg.services.analysis_starter.submitters.seqera_platform.client import SeqeraPlatformClient
from cg.services.analysis_starter.submitters.seqera_platform.dtos import (
    LaunchRequest,
    WorkflowLaunchRequest,
)
from cg.services.analysis_starter.submitters.submitter import Submitter


class SeqeraPlatformSubmitter(Submitter):

    def __init__(self, client: SeqeraPlatformClient, compute_environment_ids: dict[SlurmQos, str]):
        self.client: SeqeraPlatformClient = client
        self.compute_environment_ids: dict[str, str] = compute_environment_ids

    def submit(self, case_config: NextflowCaseConfig) -> str:
        """Starts a case and returns the workflow id for the job."""
        run_request: WorkflowLaunchRequest = self._create_launch_request(case_config)
        return self.client.run_case(run_request)

    def _create_launch_request(self, case_config: NextflowCaseConfig) -> WorkflowLaunchRequest:
        parameters: dict = read_yaml(Path(case_config.params_file))
        parameter_stream: str = write_yaml_stream(parameters)
        launch_request = LaunchRequest(
            computeEnvId=self.compute_environment_ids[case_config.case_priority],
            configProfiles=case_config.config_profiles,
            configText=f"includeConfig '{case_config.nextflow_config_file}'",
            paramsText=parameter_stream,
            pipeline=case_config.pipeline_repository,
            preRunScript=case_config.pre_run_script,
            revision=case_config.revision,
            runName=case_config.case_id,
            workDir=case_config.work_dir,
        )
        return WorkflowLaunchRequest(launch=launch_request)



"""Handler to delete data objects."""

from cg.store.crud.read import ReadHandler
from cg.store.models import Case, OrderTypeApplication, Sample


class DeleteMixin(ReadHandler):
    """Contains methods to delete business data model instances."""

    def delete_cases_without_samples(self, case_internal_ids: list[str]) -> None:
        """Delete any cases specified in case_ids without samples."""
        for case_internal_id in case_internal_ids:
            case: Case = self.get_case_by_internal_id(internal_id=case_internal_id)
            if case and not case.links:
                self.delete_item_from_store(case)
        self.commit_to_store()

    def delete_illumina_flow_cell(self, internal_id: str):
        """Delete an Illumina flow cell."""

        if flow_cell := self.get_illumina_flow_cell_by_internal_id(internal_id=internal_id):
            self.delete_item_from_store(flow_cell)
            self.commit_to_store()

    def decouple_sample_from_cases(self, sample_id: int) -> None:
        sample: Sample = self.get_sample_by_entry_id(sample_id)
        for case_sample in sample.links:
            self.delete_item_from_store(case_sample)
        self.commit_to_store()

    def delete_order_type_applications_by_application_id(self, application_id: int) -> None:
        self._get_query(OrderTypeApplication).filter_by(application_id=application_id).delete()



from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.services.deliver_files.file_fetcher.analysis_service import (
    AnalysisDeliveryFileFetcher,
)
from cg.services.deliver_files.file_fetcher.abstract import (
    FetchDeliveryFilesService,
)
from cg.services.deliver_files.file_fetcher.models import (
    DeliveryFiles,
    DeliveryMetaData,
)
from cg.services.deliver_files.file_fetcher.raw_data_service import (
    RawDataDeliveryFileFetcher,
)
from cg.services.deliver_files.tag_fetcher.abstract import (
    FetchDeliveryFileTagsService,
)
from cg.store.models import Case
from cg.store.store import Store


class RawDataAndAnalysisDeliveryFileFetcher(FetchDeliveryFilesService):

    def __init__(
        self, status_db: Store, hk_api: HousekeeperAPI, tags_fetcher: FetchDeliveryFileTagsService
    ):
        self.status_db = status_db
        self.hk_api = hk_api
        self.tags_fetcher = tags_fetcher

    def get_files_to_deliver(self, case_id: str, sample_id: str | None = None) -> DeliveryFiles:
        """
        Get files to deliver for a case or sample for both analysis and raw data.
        args:
            case_id: The case id to deliver files for
            sample_id: The sample id to deliver files for
        """
        case: Case = self.status_db.get_case_by_internal_id(internal_id=case_id)
        fastq_files: DeliveryFiles = self._fetch_files(
            service_class=RawDataDeliveryFileFetcher, case_id=case_id, sample_id=sample_id
        )
        analysis_files: DeliveryFiles = self._fetch_files(
            service_class=AnalysisDeliveryFileFetcher, case_id=case_id, sample_id=sample_id
        )
        delivery_data = DeliveryMetaData(
            case_id=case.internal_id,
            customer_internal_id=case.customer.internal_id,
            ticket_id=case.latest_ticket,
        )

        return DeliveryFiles(
            delivery_data=delivery_data,
            case_files=analysis_files.case_files,
            sample_files=analysis_files.sample_files + fastq_files.sample_files,
        )

    def _fetch_files(
        self, service_class: type, case_id: str, sample_id: str | None
    ) -> DeliveryFiles:
        """Fetch files using the provided service class.
        Wrapper to fetch files using the provided service class. This is either the RawDataDeliveryFileFetcher or the AnalysisDeliveryFileFetcher.
        args:
            service_class: The service class to use to fetch the files
            case_id: The case id to fetch files for
            sample_id: The sample id to fetch files for
        """
        service = service_class(self.status_db, self.hk_api, tags_fetcher=self.tags_fetcher)
        return service.get_files_to_deliver(case_id=case_id, sample_id=sample_id)



from pydantic import BaseModel


class DeliveryFileTags(BaseModel):
    """
    Model to hold the tags for the files to deliver.
    case_tags: The tags for the case files to deliver
    sample_tags: The tags for the sample files to deliver
    """

    case_tags: list[set[str]] | None
    sample_tags: list[set[str]]



"""Interactions with the genotype tool"""

import logging

from cg.utils.commands import Process

LOG = logging.getLogger(__name__)


class GenotypeAPI:
    """Interface with Genotype app.

    The config should contain a 'genotype' key:

        { 'database': 'mysql://localhost:3306/database' }
    """

    def __init__(self, config: dict):
        self.process = Process(
            binary=config["genotype"]["binary_path"], config=config["genotype"]["config_path"]
        )
        self.dry_run = False

    def set_dry_run(self, dry_run: bool) -> None:
        """Set the dry run state"""
        self.dry_run = dry_run

    def upload(self, bcf_path: str, samples_sex: dict, force: bool = False) -> None:
        """Upload genotypes for a family of samples."""
        upload_parameters = ["load", str(bcf_path)]
        if force:
            upload_parameters.append("--force")

        LOG.info(f"loading VCF genotypes for sample(s): {', '.join(samples_sex.keys())}")
        self.process.run_command(parameters=upload_parameters, dry_run=self.dry_run)

        for sample_id in samples_sex:
            # This is the sample sex specified by the customer
            sample_sex = samples_sex[sample_id]["pedigree"]
            self.update_sample_sex(sample_id, sample_sex)
            # This is the predicted sex based on variant calls from the workflow
            analysis_predicted_sex = samples_sex[sample_id]["analysis"]
            self.update_analysis_sex(sample_id, sex=analysis_predicted_sex)

    def update_sample_sex(self, sample_id: str, sex: str) -> None:
        """Update the sex for a sample in the genotype tool"""
        sample_sex_parameters = ["add-sex", sample_id, "-s", sex]
        LOG.debug(f"Set sex for sample {sample_id} to {sex}")
        self.process.run_command(parameters=sample_sex_parameters, dry_run=self.dry_run)

    def update_analysis_sex(self, sample_id: str, sex: str) -> None:
        """Update the predicted sex for a sample based on genotype analysis in the genotype tool"""
        analysis_sex_parameters = ["add-sex", sample_id, "-a", "sequence", sex]
        LOG.debug(f"Set predicted sex for sample {sample_id} to {sex} for the sequence analysis")
        self.process.run_command(parameters=analysis_sex_parameters, dry_run=self.dry_run)

    def __str__(self):
        return f"GenotypeAPI(dry_run: {self.dry_run})"



"""Data classes relevant to the sequencing runs."""

from pydantic import BaseModel


class RunData(BaseModel):
    """Holds information for a sequencing run."""

    pass


class RunMetrics(BaseModel):
    """Holds the metrics for a sequencing run."""

    pass


class PostProcessingDTOs(BaseModel):
    """Data transfer objects for the post-processing service."""

    pass



from http import HTTPStatus
from pathlib import Path

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from cg.clients.freshdesk.constants import EndPoints
from cg.clients.freshdesk.models import ReplyCreate, TicketCreate, TicketResponse
from cg.clients.freshdesk.utils import handle_client_errors, prepare_attachments


class FreshdeskClient:
    """Client for communicating with the freshdesk REST API."""

    def __init__(self, base_url: str, api_key: str):
        self.base_url = base_url
        self.api_key = api_key
        self.session = self._get_session()

    @handle_client_errors
    def create_ticket(self, ticket: TicketCreate, attachments: list[Path] = None) -> TicketResponse:
        """Create a ticket with multipart form data."""
        multipart_data = ticket.to_multipart_data()
        files = prepare_attachments(attachments) if attachments else None

        response = self.session.post(
            url=f"{self.base_url}{EndPoints.TICKETS}", data=multipart_data, files=files
        )
        response.raise_for_status()
        return TicketResponse.model_validate(response.json())

    @handle_client_errors
    def _get_session(self) -> Session:
        session = Session()
        session.auth = (self.api_key, "X")
        self._configure_retries(session)
        return session

    @staticmethod
    def _configure_retries(session: Session) -> None:
        """Configures retries for the session."""
        retry_strategy = Retry(
            total=5,
            status_forcelist=[
                HTTPStatus.TOO_MANY_REQUESTS,
                HTTPStatus.INTERNAL_SERVER_ERROR,
                HTTPStatus.BAD_GATEWAY,
                HTTPStatus.SERVICE_UNAVAILABLE,
                HTTPStatus.GATEWAY_TIMEOUT,
            ],
            backoff_factor=2,
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session.mount("https://", adapter)

    @handle_client_errors
    def reply_to_ticket(self, reply: ReplyCreate, attachments: list[Path] = None) -> None:
        """Send a reply to an existing ticket in Freshdesk."""
        url = f"{self.base_url}{EndPoints.TICKETS}/{reply.ticket_number}/reply"

        files = prepare_attachments(attachments) if attachments else None
        multipart_data = reply.to_multipart_data()

        response = self.session.post(url=url, data=multipart_data, files=files)
        response.raise_for_status()



"""
This module handles concatenation of balsamic fastq files.

Classes:
    FastqFileNameCreator: Creates valid balsamic filenames
    FastqHandler: Handles fastq file linking
"""

import datetime as dt
import gzip
import logging
import os
import re
import shutil
from pathlib import Path

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import FileExtensions, SequencingFileTag
from cg.constants.constants import FileFormat
from cg.io.gzip import read_gzip_first_line
from cg.models.fastq import FastqFileMeta, GetFastqFileMeta
from cg.store.models import Case, Sample
from cg.store.store import Store

LOG = logging.getLogger(__name__)

DEFAULT_DATE_STR = (
    "171015"  # Stand in value to use if sequencing date cannot be extracted from header
)
DEFAULT_INDEX = (
    "XXXXXX"  # Stand in value to use if flowcell index is to be masked when renaming file
)


def is_undetermined_in_path(file_path: Path) -> bool:
    return "Undetermined" in file_path


class FastqHandler:
    """Handles fastq file linking"""

    def __init__(self, housekeeper_api: HousekeeperAPI, root_dir: Path, status_db: Store):
        self.housekeeper_api = housekeeper_api
        self.root_dir = root_dir
        self.status_db = status_db

    def link_fastq_files(self, case_id: str) -> None:
        LOG.info(f"Linking Fastq files for case {case_id}")
        case: Case = self.status_db.get_case_by_internal_id_strict(internal_id=case_id)
        for sample in case.samples:
            fastq_dir: Path = self.get_sample_fastq_destination_dir(case=case, sample=sample)
            self.link_fastq_files_for_sample(sample=sample, fastq_dir=fastq_dir)
        LOG.info(f"Linked Fastq files for case {case_id}")

    def link_fastq_files_for_sample(self, sample: Sample, fastq_dir: Path) -> None:
        """
        Link FASTQ files for a sample to the work directory.
        """
        fastq_files: list[FastqFileMeta] = self.gather_fastq_files_for_sample(sample=sample)
        sorted_fastq_files: list[FastqFileMeta] = sorted(fastq_files, key=lambda k: k.path)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for fastq_file in sorted_fastq_files:
            fastq_file_name: str = self.create_fastq_name(
                lane=fastq_file.lane,
                flow_cell=fastq_file.flow_cell_id,
                sample=sample.internal_id,
                read_direction=fastq_file.read_direction,
                undetermined=fastq_file.undetermined,
            )
            destination_path = Path(fastq_dir, fastq_file_name)

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_file.path} -> {destination_path}")
                destination_path.symlink_to(fastq_file.path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

    def gather_fastq_files_for_sample(self, sample: Sample) -> list[FastqFileMeta]:
        return [
            self.parse_file_data(hk_file.full_path)
            for hk_file in self.housekeeper_api.files(
                bundle=sample.internal_id, tags={SequencingFileTag.FASTQ}
            )
        ]

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        raise NotImplementedError("Not implemented on parent class")

    @staticmethod
    def concatenate(files: list, concat_file: str):
        """Concatenates a list of fastq files"""
        LOG.info(FastqHandler.display_files(files, concat_file))

        with open(concat_file, "wb") as write_file_obj:
            for filename in files:
                with open(filename, "rb") as file_descriptor:
                    shutil.copyfileobj(file_descriptor, write_file_obj)

        size_before = FastqHandler.size_before(files)
        size_after = FastqHandler.size_after(concat_file)

        try:
            FastqHandler.assert_file_sizes(size_before, size_after)
        except AssertionError as error:
            LOG.warning(error)

    @staticmethod
    def size_before(files) -> int:
        """Returns the total size of the linked fastq files before concatenation"""

        return sum(os.stat(f).st_size for f in files)

    @staticmethod
    def size_after(concat_file: str) -> int:
        """returns the size of the concatenated fastq files"""

        return os.stat(concat_file).st_size

    @staticmethod
    def assert_file_sizes(size_before: int, size_after: int) -> None:
        """asserts the file sizes before and after concatenation. If the file sizes differ by more
        than 1 percent, throw an exception"""
        msg = (
            f"Warning: Large file size difference after concatenation!"
            f"Before: {size_before} -> after: {size_after}"
        )

        assert abs(size_before - size_after) / size_before <= 0.01, msg
        LOG.info("Concatenation file size check successful!")

    @staticmethod
    def display_files(files: list, concat_file: str) -> str:
        """display file names for logging purposes"""
        return (
            f"Concatenating: {', '.join(Path(file_).name for file_ in files)} -> "
            f"{Path(concat_file).name}"
        )

    @staticmethod
    def remove_files(files: list) -> None:
        for file_name in files:
            file_path: Path = Path(file_name)
            if not file_path.is_file():
                continue
            file_path.unlink()

    @staticmethod
    def get_concatenated_name(linked_fastq_name: str) -> str:
        """ "create a name for the concatenated file for some read files"""
        return f"concatenated_{'_'.join(linked_fastq_name.split('_')[-4:])}"

    @staticmethod
    def parse_fastq_header(line: str) -> FastqFileMeta | None:
        """Parse and return fastq header metadata.
        Handle Illumina's two different header formats
        @see https://en.wikipedia.org/wiki/FASTQ_format
        Raise:
            TypeError if unable to split line into expected parts.
        """
        parts = line.split(":")
        try:
            return GetFastqFileMeta.header_format.get(len(parts))(parts=parts)
        except TypeError as exception:
            LOG.error(f"Could not parse header format for header: {line}")
            raise exception

    @staticmethod
    def parse_file_data(fastq_path: Path) -> FastqFileMeta:
        header_line: str = read_gzip_first_line(file_path=fastq_path)
        fastq_file_meta: FastqFileMeta = FastqHandler.parse_fastq_header(header_line)
        fastq_file_meta.path = fastq_path
        fastq_file_meta.undetermined = is_undetermined_in_path(fastq_path)
        matches = re.findall(r"-l[1-9]t([1-9]{2})_", str(fastq_path))
        if len(matches) > 0:
            fastq_file_meta.flow_cell_id = f"{fastq_file_meta.flow_cell_id}-{matches[0]}"
        return fastq_file_meta

    @staticmethod
    def create_fastq_name(
        lane: int,
        flow_cell: str,
        sample: str,
        read_direction: int,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: str | None = None,
        meta: str | None = None,
    ) -> str:
        """Name a FASTQ file with standard conventions and
        no naming constrains from the workflow."""
        flow_cell: str = f"{flow_cell}-undetermined" if undetermined else flow_cell
        date: str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date}_{flow_cell}_{sample}_{index}_{read_direction}{FileExtensions.FASTQ}{FileExtensions.GZIP}"


class BalsamicFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: int,
        flow_cell: str,
        sample: str,
        read_direction: int,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: str | None = None,
        meta: str | None = None,
    ) -> str:
        """Name a FASTQ file following Balsamic conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        flow_cell = f"{flow_cell}-undetermined" if undetermined else flow_cell
        date: str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date}_{flow_cell}_{sample}_{index}_R_{read_direction}{FileExtensions.FASTQ}{FileExtensions.GZIP}"


class MipFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: int,
        flow_cell: str,
        sample: str,
        read_direction: int,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: str | None = None,
        meta: str | None = None,
    ) -> str:
        """Name a FASTQ file following MIP conventions."""
        flow_cell = f"{flow_cell}-undetermined" if undetermined else flow_cell
        date: str = date if isinstance(date, str) else date.strftime("%y%m%d")
        return f"{lane}_{date}_{flow_cell}_{sample}_{index}_{read_direction}{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        """Return the path to the FASTQ destination directory."""
        return Path(
            self.root_dir,
            case.internal_id,
            sample.prep_category,
            sample.internal_id,
            FileFormat.FASTQ,
        )


class MicrosaltFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: int,
        flow_cell: str,
        sample: str,
        read_direction: int,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: str | None = None,
        meta: str | None = None,
    ) -> str:
        """Name a FASTQ file following usalt conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        flow_cell = f"{flow_cell}-undetermined" if undetermined else flow_cell
        return f"{sample}_{flow_cell}_L{lane}_{read_direction}{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    def get_case_fastq_path(self, case_id: str) -> Path:
        """Get fastq paths for a case."""
        return Path(self.root_dir, "fastq", case_id)

    def get_sample_fastq_destination_dir(self, case: Case, sample: Sample) -> Path:
        return Path(self.get_case_fastq_path(case.internal_id), sample.internal_id)


class MutantFastqHandler(FastqHandler):
    @staticmethod
    def create_fastq_name(
        lane: int,
        flow_cell: str,
        sample: str,
        read_direction: int,
        date: dt.datetime = DEFAULT_DATE_STR,
        index: str = DEFAULT_INDEX,
        undetermined: str | None = None,
        meta: str | None = None,
    ) -> str:
        """Name a FASTQ file following mutant conventions. Naming must be
        xxx_R_1.fastq.gz and xxx_R_2.fastq.gz"""
        return f"{flow_cell}_L{lane}_{meta}_{read_direction}{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    @staticmethod
    def get_concatenated_name(linked_fastq_name: str) -> str:
        """ "Create a name for the concatenated file from multiple lanes"""
        return "_".join(linked_fastq_name.split("_")[2:])

    @staticmethod
    def get_nanopore_header_info(line: str) -> dict:
        fastq_meta = {"flowcell": None}
        header_metadata: list = line.split(" ")
        flowcell = header_metadata[5].split("=")
        fastq_meta["flowcell"] = flowcell[1]
        return fastq_meta

    @staticmethod
    def create_nanopore_fastq_name(
        flowcell: str,
        sample: str,
        filenr: str,
        meta: str | None = None,
    ) -> str:
        return f"{flowcell}_{sample}_{meta}_{filenr}{FileExtensions.FASTQ}{FileExtensions.GZIP}"

    @staticmethod
    def parse_nanopore_file_data(fastq_path: Path) -> dict:
        with gzip.open(fastq_path) as handle:
            header_line = handle.readline().decode()
            header_info: dict = MutantFastqHandler.get_nanopore_header_info(line=header_line)
            data = {
                "path": fastq_path,
                "flowcell": header_info["flowcell"],
            }
            return data



"""Module for Nallo Analysis API."""

import logging
from itertools import permutations
from pathlib import Path
from typing import Any

from click import File

from cg.clients.chanjo2.models import (
    CoverageMetrics,
    CoveragePostRequest,
    CoveragePostResponse,
    CoverageSample,
)
from cg.constants import Workflow
from cg.constants.constants import FileFormat, GenomeVersion
from cg.constants.nf_analysis import (
    NALLO_COVERAGE_FILE_TAGS,
    NALLO_COVERAGE_INTERVAL_TYPE,
    NALLO_COVERAGE_THRESHOLD,
    NALLO_GENERAL_METRIC_CONDITIONS,
    NALLO_PARENT_PEDDY_METRIC_CONDITION,
    NALLO_RAW_METRIC_CONDITIONS,
)
from cg.constants.scout import NALLO_CASE_TAGS, ScoutExportFileName
from cg.constants.subject import PlinkPhenotypeStatus, PlinkSex
from cg.io.controller import WriteFile
from cg.meta.workflow.nf_analysis import NfAnalysisAPI
from cg.models.analysis import NextflowAnalysis
from cg.models.cg_config import CGConfig
from cg.models.deliverables.metric_deliverables import MetricsBase, MultiqcDataJson
from cg.models.nallo.nallo import (
    NalloParameters,
    NalloQCMetrics,
    NalloSampleSheetEntry,
    NalloSampleSheetHeaders,
)
from cg.resources import NALLO_BUNDLE_FILENAMES_PATH
from cg.store.models import CaseSample, Sample

LOG = logging.getLogger(__name__)


class NalloAnalysisAPI(NfAnalysisAPI):
    """Handles communication between Nallo processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        workflow: Workflow = Workflow.NALLO,
    ):
        super().__init__(config=config, workflow=workflow)
        self.root_dir: str = config.nallo.root
        self.workflow_bin_path: str = config.nallo.workflow_bin_path
        self.profile: str = config.nallo.profile
        self.conda_env: str = config.nallo.conda_env
        self.conda_binary: str = config.nallo.conda_binary
        self.platform: str = config.nallo.platform
        self.params: str = config.nallo.params
        self.workflow_config_path: str = config.nallo.config
        self.resources: str = config.nallo.resources
        self.tower_binary_path: str = config.tower_binary_path
        self.tower_workflow: str = config.nallo.tower_workflow
        self.account: str = config.nallo.slurm.account
        self.email: str = config.nallo.slurm.mail_user
        self.compute_env_base: str = config.nallo.compute_env
        self.revision: str = config.nallo.revision
        self.nextflow_binary_path: str = config.nallo.binary_path

    @property
    def sample_sheet_headers(self) -> list[str]:
        """Headers for sample sheet."""
        return NalloSampleSheetHeaders.list()

    def get_sample_sheet_content_per_sample(self, case_sample: CaseSample) -> list[list[str]]:
        """Collect and format information required to build a sample sheet for a single sample."""
        read_file_paths = self.get_bam_read_file_paths(sample=case_sample.sample)
        sample_sheet_entries = []

        for bam_path in read_file_paths:
            sample_sheet_entry = NalloSampleSheetEntry(
                project=case_sample.case.internal_id,
                sample=case_sample.sample.internal_id,
                read_file=Path(bam_path),
                family_id=case_sample.case.internal_id,
                paternal_id=case_sample.get_paternal_sample_id or "0",
                maternal_id=case_sample.get_maternal_sample_id or "0",
                sex=self.get_sex_code(case_sample.sample.sex),
                phenotype=self.get_phenotype_code(case_sample.status),
            )
            sample_sheet_entries.extend(sample_sheet_entry.reformat_sample_content)
        return sample_sheet_entries

    @staticmethod
    def get_phenotype_code(phenotype: str) -> int:
        """Return Nallo phenotype code."""
        LOG.debug("Translate phenotype to integer code")
        try:
            code = PlinkPhenotypeStatus[phenotype.upper()]
        except KeyError:
            raise ValueError(f"{phenotype} is not a valid phenotype")
        return code

    @staticmethod
    def get_sex_code(sex: str) -> int:
        """Return Nallo sex code."""
        LOG.debug("Translate sex to integer code")
        try:
            code = PlinkSex[sex.upper()]
        except KeyError:
            raise ValueError(f"{sex} is not a valid sex")
        return code

    def get_built_workflow_parameters(self, case_id: str, dry_run: bool = False) -> NalloParameters:
        """Return parameters."""
        outdir = self.get_case_path(case_id=case_id)

        return NalloParameters(
            input=self.get_sample_sheet_path(case_id=case_id),
            outdir=outdir,
            filter_variants_hgnc_ids=f"{outdir}/{ScoutExportFileName.PANELS_TSV}",
        )

    @property
    def is_gene_panel_required(self) -> bool:
        """Return True if a gene panel needs to be created using information in StatusDB and exporting it from Scout."""
        return True

    def create_gene_panel(self, case_id: str, dry_run: bool) -> None:
        """Create and write an aggregated gene panel file exported from Scout as tsv file."""
        LOG.info("Creating gene panel file")
        bed_lines: list[str] = self.get_gene_panel(case_id=case_id, dry_run=dry_run)
        if dry_run:
            bed_lines: str = "\n".join(bed_lines)
            LOG.debug(f"{bed_lines}")
            return
        self.write_panel_as_tsv(case_id=case_id, content=bed_lines)

    def write_panel_as_tsv(self, case_id: str, content: list[str]) -> None:
        """Write the gene panel to case dir."""
        self._write_panel_as_tsv(out_dir=Path(self.root, case_id), content=content)

    @staticmethod
    def _write_panel_as_tsv(out_dir: Path, content: list[str]) -> None:
        """Write the gene panel to case dir while omitted the commented BED lines."""
        filtered_content = [line for line in content if not line.startswith("##")]
        out_dir.mkdir(parents=True, exist_ok=True)
        WriteFile.write_file_from_content(
            content="\n".join(filtered_content),
            file_format=FileFormat.TXT,
            file_path=Path(out_dir, ScoutExportFileName.PANELS_TSV),
        )

    def get_genome_build(self, case_id: str) -> GenomeVersion:
        """Return reference genome for a Nallo case. Currently fixed for hg38."""
        return GenomeVersion.HG38

    @staticmethod
    def get_bundle_filenames_path() -> Path:
        """Return Nallo bundle filenames path."""
        return NALLO_BUNDLE_FILENAMES_PATH

    def get_workflow_metrics(self, sample_id: str) -> dict:
        """Return Nallo workflow metric conditions for a sample."""
        sample: Sample = self.status_db.get_sample_by_internal_id(internal_id=sample_id)
        if "-" not in sample_id:
            metric_conditions: dict[str, dict[str, Any]] = NALLO_GENERAL_METRIC_CONDITIONS.copy()
            metric_conditions.update(NALLO_RAW_METRIC_CONDITIONS)
            self.set_somalier_sex_for_sample(sample=sample, metric_conditions=metric_conditions)
        else:
            metric_conditions = NALLO_PARENT_PEDDY_METRIC_CONDITION.copy()
        return metric_conditions

    def _get_nallo_sample_pair_patterns(self, case_id: str) -> list[str]:
        """Return sample-pair patterns for searching in MultiQC."""
        sample_ids: list[str] = list(self.status_db.get_sample_ids_by_case_id(case_id=case_id))
        pairwise_patterns: list[str] = [
            f"{sample1}-{sample2}" for sample1, sample2 in permutations(sample_ids, 2)
        ]
        return pairwise_patterns

    def get_nallo_parent_error_ped_check_metric(
        self, pair_sample_ids: str, multiqc_raw_data: dict[dict]
    ) -> MetricsBase | None:
        """Return the parsed metrics for pedigree error given a concatenated pair of sample ids."""
        metric_name: str = "parent_error_ped_check"
        peddy_metrics: dict[str, dict] = multiqc_raw_data["multiqc_peddy"]
        if sample_pair_metrics := peddy_metrics.get(pair_sample_ids, None):
            return self.get_multiqc_metric(
                metric_name=metric_name,
                metric_value=sample_pair_metrics[metric_name],
                metric_id=pair_sample_ids,
            )

    def get_nallo_raw_metric(self, sample_id: str, multiqc_raw_data: dict) -> MetricsBase | None:
        metric_name = "sex"
        raw_metrics_section: dict[str, dict] = multiqc_raw_data.get("multiqc_somalier", {})
        sample_metrics = raw_metrics_section.get(sample_id)
        if sample_metrics and metric_name in sample_metrics:
            return self.get_multiqc_metric(
                metric_name="somalier_sex",
                metric_value=sample_metrics[metric_name],
                metric_id=sample_id,
            )

    def get_nallo_multiqc_json_metrics(self, case_id: str) -> list[MetricsBase]:
        """Return a list of the Nallo metrics specified in a MultiQC json file."""
        multiqc_json: MultiqcDataJson = self.get_multiqc_data_json(case_id=case_id)
        metrics = []

        for search_pattern, metric_id in self.get_multiqc_search_patterns(case_id).items():
            metrics_for_pattern: list[MetricsBase] = (
                self.get_metrics_from_multiqc_json_with_pattern(
                    search_pattern=search_pattern,
                    multiqc_json=multiqc_json,
                    metric_id=metric_id,
                    exact_match=self.is_multiqc_pattern_search_exact,
                )
            )
            metrics.extend(metrics_for_pattern)
        for sample_id in self.status_db.get_sample_ids_by_case_id(case_id):
            if raw_metric := self.get_nallo_raw_metric(
                sample_id=sample_id,
                multiqc_raw_data=multiqc_json.report_saved_raw_data,
            ):
                metrics.append(raw_metric)
        for sample_pair in self._get_nallo_sample_pair_patterns(case_id):
            if parent_error_metric := self.get_nallo_parent_error_ped_check_metric(
                pair_sample_ids=sample_pair, multiqc_raw_data=multiqc_json.report_saved_raw_data
            ):
                metrics.append(parent_error_metric)
        metrics = self.get_deduplicated_metrics(metrics=metrics)
        return metrics

    @staticmethod
    def set_somalier_sex_for_sample(sample: Sample, metric_conditions: dict) -> None:
        if "somalier_sex" in metric_conditions:
            metric_conditions["somalier_sex"]["threshold"] = float(
                {
                    "male": PlinkSex.MALE,
                    "female": PlinkSex.FEMALE,
                    "unknown": PlinkSex.UNKNOWN,
                }[sample.sex]
            )

    def get_sample_coverage_file_path(self, bundle_name: str, sample_id: str) -> str | None:
        """Return the Nallo d4 coverage file path."""
        nallo_coverage_file_tags: list[str] = NALLO_COVERAGE_FILE_TAGS + [sample_id]
        nallo_coverage_file: File | None = self.housekeeper_api.get_file_from_latest_version(
            bundle_name=bundle_name, tags=nallo_coverage_file_tags
        )
        if nallo_coverage_file:
            return nallo_coverage_file.full_path
        LOG.warning(f"No Nallo coverage file found with the tags: {nallo_coverage_file_tags}")
        return None

    def get_sample_coverage(
        self, case_id: str, sample_id: str, gene_ids: list[int]
    ) -> CoverageMetrics | None:
        """Return sample coverage metrics from Chanjo2."""
        nallo_genome_version: GenomeVersion = self.get_genome_build(case_id)
        nallo_coverage_file_path: str | None = self.get_sample_coverage_file_path(
            bundle_name=case_id, sample_id=sample_id
        )
        try:
            post_request = CoveragePostRequest(
                build=self.translate_genome_reference(nallo_genome_version),
                coverage_threshold=NALLO_COVERAGE_THRESHOLD,
                hgnc_gene_ids=gene_ids,
                interval_type=NALLO_COVERAGE_INTERVAL_TYPE,
                samples=[
                    CoverageSample(coverage_file_path=nallo_coverage_file_path, name=sample_id)
                ],
            )
            post_response: CoveragePostResponse = self.chanjo2_api.get_coverage(post_request)
            return post_response.get_sample_coverage_metrics(sample_id)
        except Exception as error:
            LOG.error(f"Error getting coverage for sample '{sample_id}', error: {error}")
            return None

    def get_scout_upload_case_tags(self) -> dict:
        """Return Nallo Scout upload case tags."""
        return NALLO_CASE_TAGS

    def parse_analysis(self, qc_metrics_raw: list[MetricsBase], **kwargs) -> NextflowAnalysis:
        """Parse Nextflow output analysis files and return an analysis model."""
        qc_metrics_model = NalloQCMetrics
        return super().parse_analysis(
            qc_metrics_raw=qc_metrics_raw, qc_metrics_model=qc_metrics_model, **kwargs
        )

    def create_metrics_deliverables_content(self, case_id: str) -> dict[str, list[dict[str, Any]]]:
        """Create the content of a Nallo metrics deliverables file."""
        metrics: list[MetricsBase] = self.get_nallo_multiqc_json_metrics(case_id=case_id)
        self.ensure_mandatory_metrics_present(metrics=metrics)
        return {"metrics": [metric.dict() for metric in metrics]}



import logging
from functools import wraps

from cg.exc import HousekeeperBundleVersionMissingError
from cg.store.models import Case

LOG = logging.getLogger(__name__)


def handle_missing_bundle_errors(func):
    """
    Log an error when a Housekeeper bundle version is missing.

    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        case_id: str = kwargs.get("case_id")
        sample_id: str = kwargs.get("sample_id")
        case: Case = kwargs.get("case")
        try:
            return func(*args, **kwargs)
        except HousekeeperBundleVersionMissingError:
            msg: str = "Missing bundles detected for:"
            if case_id:
                msg += f" case {case_id}"
            if sample_id:
                msg += f" sample {sample_id}"
            if case:
                msg += f" case {case.internal_id}"
            LOG.error(msg)
            return []

    return wrapper



"""CLI support to create config and/or start RAREDISEASE."""

import logging
from typing import cast

import rich_click as click

from cg.cli.utils import CLICK_CONTEXT_SETTINGS, echo_lines
from cg.cli.workflow.commands import ARGUMENT_CASE_ID, resolve_compression
from cg.cli.workflow.nf_analysis import (
    config_case,
    metrics_deliver,
    report_deliver,
    run,
    start,
    start_available,
    store,
    store_available,
    store_housekeeper,
)
from cg.constants.cli_options import DRY_RUN
from cg.constants.constants import MetaApis, Workflow
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.raredisease import RarediseaseAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.services.analysis_starter.configurator.implementations.nextflow import NextflowConfigurator
from cg.services.analysis_starter.factories.configurator_factory import ConfiguratorFactory
from cg.services.analysis_starter.factories.starter_factory import AnalysisStarterFactory
from cg.services.analysis_starter.service import AnalysisStarter

LOG = logging.getLogger(__name__)


@click.group(invoke_without_command=True, context_settings=CLICK_CONTEXT_SETTINGS)
@click.pass_context
def raredisease(context: click.Context) -> None:
    """NF-core/raredisease analysis workflow."""
    AnalysisAPI.get_help(context)
    context.obj.meta_apis[MetaApis.ANALYSIS_API] = RarediseaseAnalysisAPI(config=context.obj)


raredisease.add_command(metrics_deliver)
raredisease.add_command(resolve_compression)
raredisease.add_command(config_case)
raredisease.add_command(report_deliver)
raredisease.add_command(run)
raredisease.add_command(start)
raredisease.add_command(start_available)
raredisease.add_command(store)
raredisease.add_command(store_available)
raredisease.add_command(store_housekeeper)


@raredisease.command()
@ARGUMENT_CASE_ID
@click.pass_obj
def dev_config_case(cg_config: CGConfig, case_id: str):
    """Configure a raredisease case so that it is ready to be run."""
    factory = ConfiguratorFactory(cg_config)
    configurator = cast(NextflowConfigurator, factory.get_configurator(Workflow.RAREDISEASE))
    configurator.configure(case_id=case_id)


@raredisease.command()
@ARGUMENT_CASE_ID
@click.pass_obj
def dev_run(cg_config: CGConfig, case_id: str):
    """Run a preconfigured raredisease case."""
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.RAREDISEASE
    )
    analysis_starter.run(case_id=case_id)


@raredisease.command()
@ARGUMENT_CASE_ID
@click.pass_obj
def dev_start(cg_config: CGConfig, case_id: str):
    """Start a raredisease case. Configures the case if needed."""
    factory = AnalysisStarterFactory(cg_config)
    analysis_starter: AnalysisStarter = factory.get_analysis_starter_for_workflow(
        Workflow.RAREDISEASE
    )
    analysis_starter.start(case_id=case_id)


@raredisease.command()
@click.pass_obj
def dev_start_available(cg_config: CGConfig) -> None:
    """Starts all available raredisease cases."""
    LOG.info("Starting raredisease workflow for all available cases.")
    analysis_starter = AnalysisStarterFactory(cg_config).get_analysis_starter_for_workflow(
        Workflow.RAREDISEASE
    )
    succeeded: bool = analysis_starter.start_available()
    if not succeeded:
        raise click.Abort


@raredisease.command("panel")
@DRY_RUN
@ARGUMENT_CASE_ID
@click.pass_obj
def panel(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Write aggregated gene panel file exported from Scout."""

    analysis_api: RarediseaseAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    bed_lines: list[str] = analysis_api.get_gene_panel(case_id=case_id)
    if dry_run:
        echo_lines(lines=bed_lines)
        return
    analysis_api.write_panel(case_id=case_id, content=bed_lines)


@raredisease.command("managed-variants")
@DRY_RUN
@ARGUMENT_CASE_ID
@click.pass_obj
def managed_variants(context: CGConfig, case_id: str, dry_run: bool) -> None:
    """Write managed variants file exported from Scout."""

    analysis_api: RarediseaseAnalysisAPI = context.meta_apis["analysis_api"]
    analysis_api.status_db.verify_case_exists(case_internal_id=case_id)

    vcf_lines: list[str] = analysis_api.get_managed_variants(case_id=case_id)
    if dry_run:
        echo_lines(lines=vcf_lines)
        return
    analysis_api.write_managed_variants(case_id=case_id, content=vcf_lines)
